import logging
import pickle  # noqa: 403
from abc import abstractmethod
from contextlib import contextmanager
from datetime import timedelta
from time import perf_counter
from typing import Any, Dict, Iterable, List, Literal, Optional, Sequence, Set, Tuple, Union

import pandas as pd
from rics.action_level import ActionLevel
from rics.collections.dicts import InheritedKeysDict, reverse_dict
from rics.misc import tname
from rics.performance import format_seconds

from ..exceptions import ConnectionStatusError
from ..mapping import HeuristicScore, Mapper
from ..mapping.score_functions import modified_hamming
from ..offline.types import PlaceholdersTuple, PlaceholderTranslations, SourcePlaceholderTranslations
from ..types import ID, IdType, SourceType
from . import exceptions
from ._cache import CacheAccess, CacheMetadata
from ._fetcher import Fetcher
from .types import FetchInstruction, IdsToFetch


class AbstractFetcher(Fetcher[SourceType, IdType]):
    """Base class for retrieving translations from an external source.

    .. hint::

       Clear caches with :meth:`.CacheAccess.clear_all_cache_data`.

    Args:
        mapper: A :class:`.Mapper` instance used to adapt placeholder names in sources to wanted names, i.e.
            the names of the placeholders that are in the translation :class:`.Format` being used.
        allow_fetch_all: If ``False``, an error will be raised when :meth:`fetch_all` is called.
        fetch_all_unmapped_values_action: A temporary value to use for :attr:`Mapper.unmapped_values_action
            <.Mapper.unmapped_values_action>` while :meth:`fetch_all` is executing. Setting
            ``fetch_all_unmapped_values_action='raise'`` is mutually exclusive with ``selective_fetch_all=True``.
        selective_fetch_all: If ``True``, fetch only from those :attr:`~.HasSources.sources` that contain the required
            :attr:`~.HasSources.placeholders` (after mapping).
        fetch_all_cache_max_age: If given, determines validity lifetime of data cached when :func:`fetch_all`-calls are
            made. The regular ``fetch`` function will draw from this cache as well, but only ``fetch_all`` will update
            the cache. Furthermore, caching will never be used (read or write) if :attr:`online` is ``False``.
        cache_keys: A collection of hierarchical cache-key elements, see :class:`CacheMetadata`. If given, element zero
            of the `cache_keys` is added to the :attr:`logger` name for the fetcher.

    Raises:
        rics.action_level.BadActionLevelError: If `selective_fetch_all` is ``True`` and
            `fetch_all_unmapped_values_action` is ``'raise'``.
        ValueError:  If only one of `fetch_all_cache_max_age` and `cache_keys` are given.
    """

    _FETCH: Literal["FETCH"] = "FETCH"
    _FETCH_ALL: Literal["FETCH_ALL"] = "FETCH_ALL"

    def __init__(
        self,
        mapper: Mapper[str, str, SourceType] = None,
        allow_fetch_all: bool = True,
        fetch_all_unmapped_values_action: ActionLevel.ParseType = None,
        selective_fetch_all: bool = True,
        fetch_all_cache_max_age: Union[str, pd.Timedelta, timedelta] = None,
        cache_keys: Sequence[str] = None,
    ) -> None:
        self._mapper = mapper or Mapper(**self.default_mapper_kwargs())
        if self.mapper._overrides and not self.mapper.context_sensitive_overrides:  # pragma: no cover
            raise ValueError(f"Mapper must have context-sensitive overrides (type {InheritedKeysDict.__name__}).")

        self._mapping_cache: Dict[SourceType, Dict[str, Optional[str]]] = {}
        self._allow_fetch_all = allow_fetch_all
        self._active_operation: Literal["FETCH", "FETCH_ALL", None] = None

        self._fetch_all_unmapped_values_action = (
            None
            if fetch_all_unmapped_values_action is None
            else ActionLevel.verify(
                fetch_all_unmapped_values_action,
                "AbstractFetcher.fetch_all_unmapped_values_action"
                + (f" with {selective_fetch_all=}" if selective_fetch_all else ""),
                forbidden=ActionLevel.RAISE if selective_fetch_all else None,
            )
        )
        self._selective_fetch_all = selective_fetch_all

        if fetch_all_cache_max_age and not cache_keys:  # pragma: no cover
            raise ValueError("Must specify at least one cache key with 'fetch_all_cache_max_age'.")
        self._translation_cache_access: Optional[CacheAccess[SourceType, IdType]] = None
        self._fetch_all_cache_max_age = fetch_all_cache_max_age

        logger = logging.getLogger(__package__)
        mapper_logger = logging.getLogger("id_translation.mapping.placeholders")
        if cache_keys is not None:
            cache_keys = list(cache_keys)
            adder = _ExtrasAdder(config_file=cache_keys[0])
            key0 = cache_keys[0].replace(".", "-")
            logger = logger.getChild(key0)
            mapper_logger = mapper_logger.getChild(key0)
            logger.addFilter(adder)
            mapper_logger.addFilter(adder)

        self._cache_keys: Optional[List[str]] = cache_keys
        self._logger = logger
        self._mapper.logger = mapper_logger

    def map_placeholders(
        self,
        source: SourceType,
        placeholders: Iterable[str],
        candidates: Iterable[str] = None,
        clear_cache: bool = False,
    ) -> Dict[str, Optional[str]]:
        """Map `placeholder` names to the actual names seen in `source`.

        This method calls ``Mapper.apply(values=placeholders, candidates=candidates, context=source)`` using this
        fetchers :attr:`.AbstractFetcher.mapper` instance. It is assumed that names in sources rarely change, so
        mappings are cached until the fetcher is recreated or until this method is called with ``clear_cache=True``.

        Placeholder mapping caching should not be confused with ``FETCH_ALL`` data caching.

        Args:
            source: The source to map placeholders for.
            placeholders: Desired :attr:`~.Format.placeholders`.
            candidates: A subset of candidates (placeholder names) in `source` to map with `placeholders`.
            clear_cache: If ``True``, force a full remap.

        Returns:
            A dict ``{wanted_placeholder_name: actual_placeholder_name_in_source}``, where
            `actual_placeholder_name_in_source` will be ``None`` if the wanted placeholder could not be mapped to any of
            the candidates available for the source.

        Raises:
            UnknownPlaceholderError: If any of `required_placeholders` are incorrectly mapped, or not mapped at all.

        See Also:
            🔑 This is a key event method. See :ref:`key-events` for details.
        """
        start = perf_counter()

        if clear_cache or source not in self._mapping_cache:
            self._mapping_cache[source] = {}
        ans = self._mapping_cache[source]

        candidates = set(self.get_placeholders(source) if candidates is None else candidates)
        placeholders = set(placeholders).difference(ans)  # Don't remap cached mappings

        if not placeholders:
            return ans  # Nothing new to map.

        if self.logger.isEnabledFor(logging.DEBUG):
            event_key = f"{self.__class__.__name__.upper()}.MAP_PLACEHOLDERS"
            self.logger.debug(
                f"Begin wanted-to-actual placeholder mapping of {placeholders=} to actual placeholders={candidates}"
                f" for {source=}.",
                extra=dict(
                    event_key=event_key,
                    event_stage="ENTER",
                    event_title=f"{event_key}.ENTER",
                    values=list(placeholders),
                    candidates=list(candidates),
                    context=source,
                ),
            )

        dm = self.mapper.apply(placeholders, candidates, context=source)

        for actual, wanted in dm.left_to_right.items():
            ans[actual] = wanted[0]

        for not_mapped in placeholders.difference(ans):
            ans[not_mapped] = None

        if self.logger.isEnabledFor(logging.DEBUG):
            execution_time = perf_counter() - start
            self.logger.debug(
                f"Finished wanted-to-actual placeholder mapping of {placeholders=} to actual placeholders={candidates}"
                f" for {source=}: {dm.left_to_right}.",
                extra=dict(
                    event_key=event_key,
                    event_stage="EXIT",
                    event_title=f"{event_key}.EXIT",
                    execution_time=execution_time,
                    mapping=dm.left_to_right,
                    context=source,
                ),
            )

        return ans

    def id_column(self, source: SourceType, candidates: Iterable[str] = None) -> Optional[str]:
        """Return the ID column for `source`."""
        return self.map_placeholders(source, [ID], candidates=candidates)[ID]

    @property
    def mapper(self) -> Mapper[str, str, SourceType]:
        """Return the ``Mapper`` instance used for placeholder name mapping."""
        return self._mapper

    @property
    def cache_enabled(self) -> bool:
        """Return the caching status for the fetcher."""
        return bool(self.online and self._fetch_all_cache_max_age)

    def clear_cache(self, reason: str) -> None:
        if not self.cache_enabled:
            raise ValueError("Not a cached instance.")  # pragma: no cover
        self._create_cache_access().clear(reason)

    @property
    def online(self) -> bool:
        return False  # pragma: no cover

    def assert_online(self) -> None:
        """Raise an error if offline.

        Raises:
            ConnectionStatusError: If not online.
        """
        if not self.online:  # pragma: no cover
            raise ConnectionStatusError("disconnected")

    def get_placeholders(self, source: SourceType) -> List[str]:
        """Get placeholders for `source`."""
        placeholders = self.placeholders
        if source not in placeholders:
            raise exceptions.UnknownSourceError({source}, self.sources)
        return placeholders[source]

    @property
    def allow_fetch_all(self) -> bool:
        return self._allow_fetch_all

    @property
    def logger(self) -> logging.Logger:
        """Return the ``Logger`` that is used by this instance."""
        return self._logger

    @logger.setter
    def logger(self, logger: logging.Logger) -> None:
        self._logger = logger

    @contextmanager
    def _start_operation(self, operation):  # type: ignore  # noqa
        if self._active_operation:  # pragma: no cover
            raise exceptions.ConcurrentOperationError(operation, self._active_operation)

        self._active_operation = operation
        try:
            yield
        finally:
            self._active_operation = None

    def fetch(
        self,
        ids_to_fetch: Iterable[IdsToFetch[SourceType, IdType]],
        placeholders: Iterable[str] = (),
        required: Iterable[str] = (),
    ) -> SourcePlaceholderTranslations[SourceType]:
        with self._start_operation(self._FETCH):
            return {
                itf.source: self._fetch_translations(
                    itf.source, tuple(placeholders), required_placeholders=set(required), ids=itf.ids
                )[
                    0
                ]  # Second index indicates if data is from cache -- we don't care here
                for itf in ids_to_fetch
            }

    def fetch_all(
        self,
        placeholders: Iterable[str] = (),
        required: Iterable[str] = (),
    ) -> SourcePlaceholderTranslations[SourceType]:
        if not self._allow_fetch_all:
            raise exceptions.ForbiddenOperationError(self._FETCH_ALL)

        with self._start_operation(self._FETCH_ALL):
            with self._fetch_all_mapping_context():
                return self._fetch_all(tuple(placeholders), required_placeholders=set(required))

    def _fetch_all(
        self,
        placeholders: PlaceholdersTuple,
        required_placeholders: Set[str],
    ) -> SourcePlaceholderTranslations[SourceType]:
        if self._selective_fetch_all:
            # There's nothing stopping us from doing this for regular fetching. But we assume that then the user wants
            # fetching to fail if explicit IDs can't be translated as specified.
            sources = [
                source
                for source in self.sources
                if required_placeholders.issubset(self._wanted_to_actual(source, required_placeholders))
            ]
            discarded = set(self.sources).difference(sources)
            if discarded:
                self.logger.info(
                    f"Ignoring {len(discarded)} sources {discarded} since required "
                    f"placeholders {sorted(required_placeholders)} could not be mapped by {self}."
                )
        else:
            sources = self.sources

        ans = {}
        all_from_cache = True
        for source in sources:
            translations, from_cache = self._fetch_translations(source, placeholders, required_placeholders)
            ans[source] = translations
            all_from_cache = all_from_cache and from_cache

        if not all_from_cache:
            self._put_cached_translations(ans)

        return ans

    @contextmanager
    def _fetch_all_mapping_context(self):  # type: ignore  # noqa
        fetch_all_unmapped_values_action = self._fetch_all_unmapped_values_action
        selective_fetch_all = self._selective_fetch_all
        if fetch_all_unmapped_values_action is None and not selective_fetch_all:
            yield
            return

        unmapped_values_action = fetch_all_unmapped_values_action or ActionLevel.IGNORE

        if self.mapper.unmapped_values_action == unmapped_values_action:
            yield
            return

        original_mapper = self._mapper
        self._mapper = self._mapper.copy(unmapped_values_action=unmapped_values_action)
        try:
            self.logger.info(
                f"Using Mapper.{unmapped_values_action=} until the current {self._FETCH_ALL}-operation finishes, "
                f"since {selective_fetch_all=} and {fetch_all_unmapped_values_action=}."
            )
            yield
        finally:
            self._mapper = original_mapper

    @abstractmethod
    def fetch_translations(
        self, instruction: FetchInstruction[SourceType, IdType]
    ) -> PlaceholderTranslations[SourceType]:
        """Retrieve placeholder translations from the source.

        Args:
            instruction: A single instruction for IDs to fetch. If IDs is ``None``, the fetcher should retrieve data for
                as many IDs as possible.

        Returns:
            Placeholder translation elements.

        Raises:
            UnknownPlaceholderError: If the placeholder is unknown to the fetcher.

        See Also:
            🔑 This is a key event method. See :ref:`key-events` for details.
        """

    def _fetch_translations(
        self,
        source: SourceType,
        placeholders: PlaceholdersTuple,
        required_placeholders: Set[str],
        ids: Iterable[IdType] = None,
    ) -> Tuple[PlaceholderTranslations[SourceType], bool]:
        reverse_mappings, instr = self._make_fetch_instruction(source, placeholders, required_placeholders, ids)

        cached_translations = self._get_cached_translations(instr.source)
        if cached_translations is not None:
            self.logger.debug(f"Returning cached translations for {source=}.")
            return cached_translations, True

        if self.logger.isEnabledFor(logging.DEBUG):
            event_key = f"{self.__class__.__name__.upper()}.FETCH_TRANSLATIONS"
            self.logger.debug(
                f"Begin fetching {placeholders=} from {source=} for {'all' if instr.ids is None else len(instr.ids)} IDs.",
                extra=dict(
                    event_key=event_key,
                    event_stage="ENTER",
                    event_title=f"{event_key}.ENTER",
                    source=source,
                    placeholders=instr.placeholders,
                    required_placeholders=list(instr.required),
                    num_ids=None if instr.ids is None else len(instr.ids),
                    fetch_all=instr.fetch_all,
                ),
            )
        start = perf_counter()
        translations = self.fetch_translations(instr)
        if self.logger.isEnabledFor(logging.DEBUG):
            execution_time = perf_counter() - start
            self.logger.debug(
                f"Finished fetching placeholders={translations.placeholders} for {len(translations.records)} IDs "
                f"from source '{translations.source}' in {format_seconds(execution_time)} using {self}.",
                extra=dict(
                    event_key=event_key,
                    event_stage="EXIT",
                    event_title=f"{event_key}.EXIT",
                    execution_time=execution_time,
                    source=source,
                    placeholders=translations.placeholders,
                    num_ids=len(translations.records),
                    fetch_all=instr.fetch_all,
                ),
            )

        if reverse_mappings:
            # The mapping is only in reverse from the Fetchers point-of-view; we're mapping back to "proper" values.
            translations.placeholders = tuple(reverse_mappings.get(p, p) for p in translations.placeholders)

        translations.id_pos = translations.placeholders.index(ID)

        unmapped_required_placeholders = required_placeholders.difference(translations.placeholders)
        if unmapped_required_placeholders:
            self._verify_placeholders(reverse_mappings or {}, source, unmapped_required_placeholders)

        return translations, False

    def _verify_placeholders(self, reverse_mappings: Dict[str, str], source: SourceType, unmapped: Set[str]) -> None:
        hint = ""
        if unmapped.intersection(reverse_mappings.values()):
            r = reverse_dict(reverse_mappings)
            bad_mappings = {b: r[b] for b in unmapped}
            hint = (
                f"\nHint: Mapping {bad_mappings} for required placeholders (keys) were made to placeholders that do not"
                f" exist. The override configuration {self.mapper._overrides} may be incorrect."
            )
        raise exceptions.UnknownPlaceholderError(
            f"Required placeholders {unmapped} not recognized. For {source=}, known placeholders are: "
            f"{sorted(self.placeholders[source])} for {self}.{hint}"
        )

    def _make_fetch_instruction(
        self,
        source: SourceType,
        placeholders: PlaceholdersTuple,
        required_placeholders: Set[str],
        ids: Optional[Iterable[IdType]],
    ) -> Tuple[Optional[Dict[str, str]], FetchInstruction[SourceType, IdType]]:
        required_placeholders.add(ID)
        if ID not in placeholders:
            placeholders = (ID,) + placeholders

        wanted_to_actual = self._wanted_to_actual(source, placeholders)

        actual_to_wanted: Dict[str, str] = reverse_dict(wanted_to_actual)
        need_placeholder_mapping = actual_to_wanted != wanted_to_actual
        if need_placeholder_mapping:
            # We'll just map what we can here. If anything is missing it'll be caught later.
            def apply(c: Iterable[str]) -> Iterable[str]:
                return (wanted_to_actual[p] for p in c if p in wanted_to_actual)

            placeholders = tuple(apply(placeholders))
            required_placeholders = set(apply(required_placeholders))

        return (
            actual_to_wanted if need_placeholder_mapping else None,
            FetchInstruction(
                source=source,
                placeholders=placeholders,
                required=required_placeholders,
                ids=None if ids is None else list(ids),
            ),
        )

    def _wanted_to_actual(self, source: SourceType, wanted_placeholders: Iterable[str]) -> Dict[str, str]:
        wanted_to_actual = self.map_placeholders(source, wanted_placeholders)
        return {wanted: actual for wanted, actual in wanted_to_actual.items() if actual is not None}

    def _create_cache_access(self) -> CacheAccess[SourceType, IdType]:
        assert self._cache_keys is not None  # noqa: S101
        return CacheAccess(
            self._fetch_all_cache_max_age,
            CacheMetadata(self._cache_keys, placeholders=self.placeholders),
        )

    def _get_cached_translations(self, source: SourceType) -> Optional[PlaceholderTranslations[SourceType]]:
        if not self.cache_enabled:
            return None
        if self._translation_cache_access is None:
            self._translation_cache_access = self._create_cache_access()
        return self._translation_cache_access.read_cache(source)

    def _put_cached_translations(self, data: SourcePlaceholderTranslations[SourceType]) -> None:
        if not self.cache_enabled:
            return
        self._translation_cache_access = self._create_cache_access()
        self._translation_cache_access.write_cache(data)

    @classmethod
    def default_mapper_kwargs(cls) -> Dict[str, Any]:
        """Return default ``Mapper`` arguments for ``AbstractFetcher`` implementations."""
        return dict(
            score_function=HeuristicScore(
                cls.default_score_function,  # type: ignore
                heuristics=[("force_lower_case", {})],
            ),
            overrides=InheritedKeysDict(),
        )

    @classmethod
    def default_score_function(cls, value: str, candidates: Iterable[str], context: str) -> Iterable[float]:
        """Compute score for candidates."""
        return modified_hamming(value, candidates, context)

    def __str__(self) -> str:
        sources = self.sources if self.sources else "<no sources>"
        return f"{tname(self)}({sources=})"


class _ExtrasAdder(logging.Filter):
    def __init__(self, **extras: Any) -> None:
        super().__init__()
        self.extras = extras

    def filter(self, record: logging.LogRecord) -> bool:  # noqa: A003
        for name, value in self.extras.items():
            setattr(record, name, value)
        return True



import logging
import warnings
from dataclasses import dataclass
from time import perf_counter
from typing import Any, Dict, Iterable, List, Literal, Optional, Set, TypeVar
from urllib.parse import quote_plus

import sqlalchemy
from rics.misc import tname
from rics.performance import format_perf_counter

from ..offline.types import PlaceholderTranslations
from ..types import IdType
from . import exceptions, support
from ._abstract_fetcher import AbstractFetcher
from .types import FetchInstruction


class SqlFetcher(AbstractFetcher[str, IdType]):
    """Fetch data from a SQL source. Requires SQLAlchemy.

    Args:
        connection_string: A SQLAlchemy connection string.
        password: Password to insert into the connection string. Will be escaped to allow for special characters. If
            given, the connection string must contain a password key, eg; ``dialect://user:{password}@host:port``.
        whitelist_tables: The only tables the ``SqlFetcher`` may access. Mutually exclusive with `blacklist_tables`.
        blacklist_tables: The only tables the ``SqlFetcher`` may not access. Mutually exclusive with `whitelist_tables`.
        schema: Database schema to use. Typically needed only if `schema` is not the default schema for the user
            specified in the connection string.
        include_views: If ``True``, the fetcher will discover and query views as well.
        fetch_all_limit: Maximum size of table to allow a fetch all-operation. 0=never allow. Ignore if ``None``.
        engine_kwargs: A dict of keyword arguments for :func:`sqlalchemy.create_engine`.
        **kwargs: Primarily passed to ``super().__init__``, then used as :meth:`selection_filter_type` kwargs.

    Raises:
        ValueError: If both `whitelist_tables` and `blacklist_tables` are given.

    Notes:
        Inheriting classes may override on or more of the following methods to further customize operation.

        * :meth:`create_engine`; initializes the SQLAlchemy engine. Calls ``create_engine``.
        * :meth:`parse_connection_string`; does basic URL encoding.
        * :meth:`make_table_summary`; creates :class:`TableSummary` instances. Calls ``get_approximate_table_size``.
        * :meth:`get_approximate_table_size`; default is ``SELECT count(*) FROM table``.
        * :meth:`selection_filter_type`; control what kind of filtering (if any) should be done when selecting IDs.
        * :meth:`finalize_statement`; adjust the final query, e.g. to apply additional filtering.

        Overriding should be done with care, as these may call each other internally.
    """

    def __init__(
        self,
        connection_string: str,
        password: str = None,
        whitelist_tables: Iterable[str] = None,
        blacklist_tables: Iterable[str] = (),
        schema: str = None,
        include_views: bool = False,
        fetch_all_limit: Optional[int] = 100_000,
        engine_kwargs: Dict[str, Any] = None,
        **kwargs: Any,
    ) -> None:
        if kwargs:
            import inspect

            parameters = set(inspect.signature(super().__init__).parameters)
            super_kwargs = {k: kwargs.pop(k) for k in parameters.intersection(kwargs)}
            self._select_params = kwargs
        else:  # pragma: no cover
            self._select_params = {}
            super_kwargs = {}
        super().__init__(**super_kwargs)

        self._engine = self.create_engine(connection_string, password, engine_kwargs or {})
        self._estr = str(self.engine)
        self._schema = schema
        self._reflect_views = include_views
        self._fetch_all_limit = fetch_all_limit

        self._blacklist = set(blacklist_tables)

        self._table_ts_dict: Optional[Dict[str, SqlFetcher.TableSummary]] = None

        if whitelist_tables is None:
            self._whitelist = set()
        else:
            if blacklist_tables:
                raise ValueError("At most one of whitelist and blacklist may be given.")  # pragma: no cover

            whitelist_tables = set(whitelist_tables)
            if len(whitelist_tables) == 0:
                self.close()
                msg = f"Got empty 'whitelist_tables' argument. No tables will be available to {self}."
                self.logger.getChild("sql").warning(msg)
                warnings.warn(msg, stacklevel=2)

            self._whitelist = set(whitelist_tables)

    @property
    def _summaries(self) -> Dict[str, "SqlFetcher.TableSummary"]:
        """Names and sizes of tables that the ``SqlFetcher`` may interact with."""
        if self._table_ts_dict is None:
            self._table_ts_dict = self._get_summaries()

        return self._table_ts_dict

    def fetch_translations(self, instr: FetchInstruction[str, IdType]) -> PlaceholderTranslations[str]:
        """Fetch columns from a SQL database."""
        ts = self._summaries[instr.source]
        columns = ts.select_columns(instr)
        select = sqlalchemy.select(*map(ts.columns.get, columns))  # type: ignore[arg-type]

        if instr.ids is None and not ts.fetch_all_permitted:  # pragma: no cover
            raise exceptions.ForbiddenOperationError(self._FETCH_ALL, f"disabled for table '{ts.name}'.")

        stmt = select if instr.ids is None else self._make_query(ts, select, set(instr.ids))
        stmt = self.finalize_statement(stmt, ts.id_column, ts.id_column.table)

        with self.engine.connect() as conn:
            records = tuple(map(tuple, conn.execute(stmt)))
        return PlaceholderTranslations(instr.source, tuple(columns), records)

    StatementType = TypeVar("StatementType", bound=sqlalchemy.sql.Executable)
    """Input and return bounds for :meth:`finalize_statement`."""

    def finalize_statement(
        self,
        statement: StatementType,
        id_column: sqlalchemy.Column,  # type:ignore[type-arg]
        table: sqlalchemy.Table,
    ) -> StatementType:
        """Finalize a statement before execution. Does nothing by default.

        Args:
            statement: A pre-build select query.
            id_column: The ID column of `table`.
            table: Table from which the selection `select` is made.

        Returns:
            The final statement to execute.
        """
        return statement

    def _make_query(
        self,
        ts: "SqlFetcher.TableSummary",
        select: sqlalchemy.sql.Select,  # type: ignore[type-arg]
        ids: Set[IdType],
    ) -> sqlalchemy.sql.Select:  # type: ignore[type-arg]
        where = self.selection_filter_type(ids, ts, **self._select_params)

        if where == "in":
            return select.where(ts.id_column.in_(ids))
        if where == "between":
            return select.where(ts.id_column.between(min(ids), max(ids)))
        if where is None:
            return select

        raise ValueError(f"Bad response {where=} returned by {self.selection_filter_type=}.")  # pragma: no cover

    @property
    def online(self) -> bool:
        return self._engine is not None  # pragma: no cover

    @property
    def sources(self) -> List[str]:
        return list(self._summaries)

    @property
    def placeholders(self) -> Dict[str, List[str]]:
        return {name: list(ts.columns.keys()) for name, ts in self._summaries.items()}

    @property
    def allow_fetch_all(self) -> bool:
        return super().allow_fetch_all and all(s.fetch_all_permitted for s in self._summaries.values())

    def __str__(self) -> str:
        disconnected = "<disconnected>: " if not self.online else ""
        schema = f"[schema={self._schema!r}]" if self._schema else ""
        return f"{tname(self)}({disconnected}{self._estr}, tables{schema}={repr(self.sources or '<no tables>')})"

    @property
    def engine(self) -> sqlalchemy.engine.Engine:
        """Engine used by this fetcher.

        Returns:
            The ``Engine`` used for fetching.
        """
        self.assert_online()
        return self._engine

    def close(self) -> None:  # pragma: no cover
        if self._engine is None:
            return

        self.logger.getChild("sql").debug("Dispose %s", self._estr)
        self._table_ts_dict = {}
        self._engine.dispose()

    @classmethod
    def create_engine(
        cls,
        connection_string: str,
        password: Optional[str],
        engine_kwargs: Dict[str, Any],
    ) -> sqlalchemy.engine.Engine:
        """Factory method used by ``__init__``.

        For a more detailed description of the arguments and the behaviour of this function, see the
        :class:`class docstring <SqlFetcher>`.

        Args:
            connection_string: A SQLAlchemy connection string.
            password: Password to insert into the connection string.
            engine_kwargs: A dict of keyword arguments for :func:`sqlalchemy.create_engine`.

        Returns:
            A new ``Engine``.
        """
        return sqlalchemy.create_engine(
            cls.parse_connection_string(connection_string, password),
            **engine_kwargs,
        )

    @classmethod
    def parse_connection_string(cls, connection_string: str, password: Optional[str]) -> str:  # pragma: no cover
        """Parse a connection string."""
        if password:
            if "{password}" in connection_string:
                connection_string = connection_string.format(password=quote_plus(password))
            else:  # pragma: no cover
                warnings.warn(
                    "A password was specified, but the connection string does not have a {password} key.", stacklevel=3
                )
        return connection_string

    def _get_summaries(self) -> Dict[str, "SqlFetcher.TableSummary"]:
        start = perf_counter()
        metadata = self.get_metadata()

        logger = self.logger.getChild("sql").getChild("discovery")
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug(f"{self._estr}: Metadata created in {format_perf_counter(start)}.")

        table_names = {t.name for t in metadata.tables.values()}
        if self._whitelist:
            tables = self._whitelist  # pragma: no cover
        else:
            tables = table_names.difference(self._blacklist) if self._blacklist else table_names

        if not tables:  # pragma: no cover
            if self._whitelist:
                extra = f" (whitelist: {self._whitelist})"
            elif self._blacklist:
                extra = f" (blacklist: {self._blacklist})"
            else:
                extra = ""

            logger.warning(f"{self._estr}: No sources found{extra}. Available tables: {table_names}")

        ans = {}
        for name in tables:
            qualified_name = name if self._schema is None else f"{self._schema}.{name}"
            table = metadata.tables[qualified_name]
            id_column = self.id_column(table.name, (c.name for c in table.columns))
            if id_column is None:  # pragma: no cover
                continue  # Mapper would've raised an error if we required all non-filtered tables to be mapped
            if id_column not in table.columns:
                raise exceptions.UnknownPlaceholderError(
                    f"The ID column {id_column!r} is not present in table={table.name!r}. The override "
                    f"configuration {self.mapper._overrides} may be incorrect. "
                    f"Known columns: {sorted(c.name for c in table.columns)}."
                )

            ans[str(name)] = self.make_table_summary(table, table.columns[id_column])

        return ans

    def make_table_summary(
        self,
        table: sqlalchemy.Table,
        id_column: sqlalchemy.Column,  # type: ignore[type-arg]
    ) -> "SqlFetcher.TableSummary":
        """Create a table summary."""
        start = perf_counter()
        size = self.get_approximate_table_size(table, id_column)
        logger = self.logger.getChild("sql")
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug(f"{self._estr}: Counted {size} rows of table '{table.name}' in {format_perf_counter(start)}.")
        fetch_all_permitted = self._fetch_all_limit is None or size < self._fetch_all_limit
        return SqlFetcher.TableSummary(str(table.name), size, table.columns, fetch_all_permitted, id_column)

    def get_approximate_table_size(
        self,
        table: sqlalchemy.Table,
        id_column: sqlalchemy.Column,  # type: ignore[type-arg]
    ) -> int:
        """Return the approximate size of a table.

        Called only by the :meth:`make_table_summary` method during discovery. The default implementation performs a
        count on the ID column, which may be expensive.

        Args:
            table: A table object.
            id_column: The ID column in `table`.

        Returns:
            An approximate size for `table`.
        """
        stmt = sqlalchemy.func.count(id_column)
        stmt = self.finalize_statement(stmt, id_column, table)

        with self.engine.connect() as conn:
            return int(conn.execute(stmt).scalar())  # type: ignore[arg-type]

    def get_metadata(self) -> sqlalchemy.MetaData:
        """Create a populated metadata object."""
        metadata = sqlalchemy.MetaData(schema=self._schema)
        metadata.reflect(self.engine, only=tuple(self._whitelist) or None, views=self._reflect_views)
        return metadata

    @classmethod
    def selection_filter_type(
        cls,
        ids: Set[IdType],
        table_summary: "SqlFetcher.TableSummary",
        fetch_all_below: int = 25,
        fetch_all_above_ratio: float = 0.90,
        fetch_in_below: int = 1200,
        fetch_between_over: int = 10_000,
        fetch_between_max_overfetch_factor: float = 2.5,
    ) -> Literal["in", "between", None]:
        """Determine the type of filter (``WHERE``-query) to use, if any.

        In the descriptions below, ``len(table)`` refers to the :attr:`TableSummary.size`-attribute of `table_summary`.
        Bare select implies fetching the entire table.

        Args:
            ids: IDs to fetch.
            table_summary: A summary of the table that's about to be queried.
            fetch_all_below: Use bare select if ``len(ids) <= len(table)``.
            fetch_all_above_ratio: Use bare select if ``len(ids) > len(table) * ratio``.
            fetch_in_below: Always use ``IN``-clause when fetching less than `fetch_in_below` IDs.
            fetch_between_over: Always use ``BETWEEN``-clause when fetching more than `fetch_between_over` IDs.
            fetch_between_max_overfetch_factor: If number of IDs to fetch is between `fetch_in_below` and
                `fetch_between_over`, use this factor to choose between ``IN`` and ``BETWEEN`` clause.

        Returns:
            One of (``'in', 'between', None``), where ``None`` means bare select (fetch the whole table).

        Notes:
            Override this function to redefine ``SELECT`` filtering logic.
        """
        num_ids = len(ids)
        size = float("inf") if table_summary.size <= 0 else table_summary.size
        table = table_summary.name

        # Just fetch everything if we're getting "most of" the data anyway
        if size <= fetch_all_below or num_ids / size > fetch_all_above_ratio:
            if num_ids > size:  # pragma: no cover
                warnings.warn(f"Fetching {num_ids} unique IDs from {table=} which only has {size} rows.", stacklevel=2)
            return None

        if num_ids < fetch_in_below:
            return "in"

        min_id, max_id = min(ids), max(ids)

        if isinstance(next(iter(ids)), str) or num_ids > fetch_between_over:
            return "between"

        try:
            overfetch_factor = (max_id - min_id) / num_ids  # type: ignore
        except TypeError:  # pragma: no cover
            return "between"  # Non-numeric ID type

        if overfetch_factor > fetch_between_max_overfetch_factor:
            return "in"
        else:
            return "between"

    @dataclass(frozen=True)
    class TableSummary:
        """Brief description of a known table."""

        name: str
        """Name of the table."""
        size: int
        """Approximate size of the table."""
        columns: sqlalchemy.sql.ColumnCollection  # type: ignore[type-arg]
        """A flag indicating that the FETCH_ALL-operation is permitted for this table."""
        fetch_all_permitted: bool
        """A flag indicating that the FETCH_ALL-operation is permitted for this table."""
        id_column: sqlalchemy.Column  # type: ignore[type-arg]
        """The ID column of the table."""

        def select_columns(self, instr: FetchInstruction[str, IdType]) -> List[str]:
            """Return required and optional columns of the table."""
            return support.select_placeholders(instr, self.columns.keys())



"""General errors for the translation suite."""


class ConfigurationError(ValueError):
    """Raised in case of bad configuration."""


class ConnectionStatusError(ValueError):
    """Raised when trying to perform operations in a bad online/offline state."""


class TranslationError(ValueError):
    """Base class for translation errors."""


class TooManyFailedTranslationsError(TranslationError):
    """Raised if too many IDs fail to translate."""



"""Errors and warnings related to fetching."""
from typing import Any, Iterable


class FetcherWarning(RuntimeWarning):
    """Base class for ``Fetcher`` warnings."""


class FetcherError(RuntimeError):
    """Base class for ``Fetcher`` exceptions."""


class ForbiddenOperationError(FetcherError):
    """Exception indicating that the ``Fetcher`` does not support an operation."""

    def __init__(self, operation: str, reason: str = "not supported by this fetcher.") -> None:
        super().__init__(f"Operation '{operation}' " + reason)
        self.operation = operation


class ConcurrentOperationError(FetcherError):
    """Exception indicating that an operation is already in progress."""

    def __init__(self, operation: str, active_operation: str) -> None:  # pragma: no cover
        super().__init__(f"Operation '{operation}' cannot begin while a '{active_operation}'-operation is in progress.")
        self.operation = operation
        self.active_operation = active_operation


class ImplementationError(FetcherError):
    """An underlying implementation did something wrong."""


class UnknownPlaceholderError(FetcherError):
    """Caller requested unknown placeholder name(s)."""


class UnknownIdError(FetcherError):
    """Caller requested unknown id(s)."""


class UnknownSourceError(FetcherError):
    """Caller requested unknown source(s)."""

    def __init__(self, unknown_sources: Iterable[Any], sources: Iterable[Any]) -> None:
        super().__init__(f"Sources {set(unknown_sources)} not recognized: Known {sources=}.")


class DuplicateSourceWarning(FetcherWarning):
    """Duplicate sources detected."""


class DuplicateSourceError(FetcherError):
    """Multiple translations for the same source received."""



from abc import ABC, abstractmethod
from typing import Any, Dict, Generic, List

from rics.misc import tname

from ..types import ID, IdType, NameType, SourceType
from ._format import Format
from ._magic_dict import MagicDict
from .types import PlaceholdersTuple, PlaceholderTranslations, TranslatedIds


class FormatApplier(ABC, Generic[NameType, SourceType, IdType]):
    """Base class for application of ``Format`` specifications.

    Args:
        translations: Matrix of ID translation components returned by fetchers.

    Raises:
        ValueError: If `default` is given and any placeholder names are missing.
    """

    def __init__(self, translations: PlaceholderTranslations[SourceType]) -> None:
        self._source = translations.source
        self._placeholder_names = translations.placeholders
        self._n_ids = len(translations.records)

    def __call__(
        self,
        fmt: Format,
        placeholders: PlaceholdersTuple = None,
        default_fmt: Format = None,
        default_fmt_placeholders: Dict[str, Any] = None,
    ) -> MagicDict[IdType]:
        """Translate IDs.

        Args:
            fmt: Translation format to use.
            placeholders: Placeholders to include in the formatted output. Use as many as possible if ``None``.
            default_fmt: Alternative format for default translation.
            default_fmt_placeholders: Default placeholders

        Returns:
            A dict ``{idx: translated_id}``.
        """
        if placeholders is None:
            # Use as many placeholders as possible.
            placeholders = tuple(filter(self._placeholder_names.__contains__, fmt.placeholders))

        fstring = fmt.fstring(placeholders, self.positional)
        real_translations = self._apply(fstring, placeholders)

        if default_fmt or default_fmt_placeholders:
            fmap = {ID: "{}", **(default_fmt_placeholders or {})}
            default_fstring = (default_fmt or fmt).fstring(fmap, positional=False).format(**fmap)
        else:
            default_fstring = None

        return MagicDict(
            real_translations,
            default_fstring,
        )

    @abstractmethod
    def _apply(self, fstring: str, placeholders: PlaceholdersTuple) -> TranslatedIds[IdType]:
        """Apply fstring to all IDs.

        The abstract class delegates ``__apply__``-invocations to this method after some input validation.

        Args:
            fstring: A format string.
            placeholders: Keys needed for the fstring, in the order in which they appear.

        Returns:
            A dict ``{idx: translated_id}``.
        """

    @property
    def positional(self) -> bool:
        """If ``True``, names are stripped from fstring placeholders."""
        return True

    @property
    def source(self) -> SourceType:
        """Return translation source."""
        return self._source

    @property
    def placeholders(self) -> List[str]:
        """Return placeholder names in sorted order."""
        return list(self._placeholder_names)

    def __len__(self) -> int:
        return self._n_ids  # pragma: no cover

    def __repr__(self) -> str:
        placeholders = tuple(self._placeholder_names)
        source = self._source
        return f"{tname(self)}({len(self)} IDs, {placeholders=}, {source=})"


class DefaultFormatApplier(FormatApplier[NameType, SourceType, IdType]):
    """Default format applier implementation."""

    def __init__(
        self,
        translations: PlaceholderTranslations[SourceType],
    ) -> None:
        super().__init__(translations)
        self._pht = translations

    def _apply(self, fstring: str, placeholders: PlaceholdersTuple) -> TranslatedIds[IdType]:
        if self._placeholder_names == placeholders:
            return {record[self._pht.id_pos]: fstring.format(*record) for record in self._pht.records}
        else:
            pos = tuple(map(self._placeholder_names.index, placeholders))
            return {record[self._pht.id_pos]: fstring.format(*(record[i] for i in pos)) for record in self._pht.records}



import logging
import warnings
from datetime import timedelta
from pathlib import Path
from time import perf_counter
from typing import Any, Dict, Generic, Iterable, List, Optional, Set, Tuple, Type, Union

import numpy
import pandas as pd
from rics._internal_support.types import PathLikeType
from rics.collections.dicts import InheritedKeysDict, MakeType
from rics.collections.misc import as_list
from rics.misc import tname
from rics.performance import format_perf_counter, format_seconds

from ._config_utils import ConfigMetadata
from .dio import DataStructureIO, resolve_io
from .exceptions import ConnectionStatusError, TooManyFailedTranslationsError
from .factory import TranslatorFactory
from .fetching import Fetcher
from .fetching.types import IdsToFetch
from .mapping import DirectionalMapping, Mapper
from .mapping.exceptions import MappingError, MappingWarning
from .mapping.types import UserOverrideFunction
from .offline import Format, TranslationMap
from .offline.types import FormatType, PlaceholderTranslations, SourcePlaceholderTranslations
from .types import ID, HasSources, IdType, Names, NamesPredicate, NameType, NameTypes, SourceType, Translatable

_NAME_ATTRIBUTES = ("name", "columns", "keys")

LOGGER = logging.getLogger(__package__).getChild("Translator")

FetcherTypes = Union[
    TranslationMap[NameType, SourceType, IdType],
    Fetcher[SourceType, IdType],
    SourcePlaceholderTranslations[SourceType],
    Dict[SourceType, PlaceholderTranslations.MakeTypes],
]


class Translator(Generic[NameType, SourceType, IdType], HasSources[SourceType]):
    """Translate IDs to human-readable labels.

    For an introduction to translation, see the :ref:`translation-primer` page.

    The recommended way of initializing ``Translator`` instances is the :meth:`from_config` method. For configuration
    file details, please refer to the :ref:`translator-config` page.

    The `Translator` is the main entry point for all translation tasks. Simplified translation process steps:

        1. The :attr:`map` method performs name-to-source mapping (see :class:`.mapping.DirectionalMapping`).
        2. The :attr:`fetch` method extracts IDs to translate and retrieves data (see :class:`.TranslationMap`).
        3. Finally, the :attr:`translate` method applies the translations and returns to the caller.

    Args:
        fetcher: A :class:`.Fetcher` or ready-to-use translations.
        fmt: String :class:`.Format` specification for translations.
        mapper: A :class:`~.mapping.Mapper` instance for binding names to sources.
        default_fmt: Alternative :class:`.Format` to use instead of `fmt` for fallback translation of unknown IDs.
        default_fmt_placeholders: Shared and/or source-specific default placeholder values for unknown IDs. See
            :meth:`rics.collections.dicts.InheritedKeysDict.make` for details.
        allow_name_inheritance: If ``True``, enable name resolution fallback to the parent `translatable` when
            translating with the ``attribute``-option. Allows nameless ``pandas.Index`` instances to inherit the name of
            a ``pandas.Series``.

    Notes:
        Untranslatable IDs will be ``None`` by default if neither `default_fmt` nor `default_fmt_placeholders` is given.
        Adding the `maximal_untranslated_fraction` option to :meth:`translate` will raise an exceptions if too many IDs
        are left untranslated. Note however that this verifiction step may be expensive.

    Examples:
        A minimal example. For a more complete use case, see the :ref:`dvdrental` example. Assume that we have data for
        people and animals as in the table below::

            people:                       animals:
                 id | name    | gender       id | name   | is_nice
              ------+---------+--------     ----+--------+---------
               1991 | Richard | Male          0 | Tarzan | false
               1999 | Sofia   | Female        1 | Morris | true
               1904 | Fred    | Male          2 | Simba  | true

        In most real cases we'd fetch this table from somewhere. In this case, however, there's so little data that we
        can simply enumerate the components needed for translation ourselves to create a :class:`.MemoryFetcher`.

        >>> from id_translation import Translator
        >>> translation_data = {
        ...     'animals': {'id': [0, 1, 2], 'name': ['Tarzan', 'Morris', 'Simba'], 'is_nice': [False, True, True]},
        ...     'people': {'id': [1999, 1991, 1904], 'name': ['Sofia', 'Richard', 'Fred']},
        ... }
        >>> translator = Translator(translation_data, fmt='{id}:{name}[, nice={is_nice}]')
        >>> data = {'animals': [0, 2], 'people': [1991, 1999]}
        >>> for key, translated_table in translator.translate(data).items():
        >>>     print(f'Translations for {repr(key)}:')
        >>>     for translated_id in translated_table:
        >>>         print(f'    {repr(translated_id)}')
        Translations for 'animals':
            '0:Tarzan, nice=False'
            '2:Simba, nice=True'
        Translations for 'people':
            '1991:Richard'
            '1999:Sofia'

        Handling unknown IDs.

        >>> default_fmt_placeholders = dict(
        ...     default={'is_nice': 'Maybe?', 'name': "Bob"},
        ...     specific={'animals': {'name': 'Fido'}},
        >>> )
        >>> useless_database = {
        ...     'animals': {'id': [], 'name': []},
        ...     'people': {'id': [], 'name': []}
        >>> }
        >>> translator = Translator(useless_database, default_fmt_placeholders=default_fmt_placeholders,
        ...                         fmt='{id}:{name}[, nice={is_nice}]')
        >>> data = {'animals': [0], 'people': [0]}
        >>> for key, translated_table in translator.translate(data).items():
        >>>     print(f'Translations for {repr(key)}:')
        >>>     for translated_id in translated_table:
        >>>         print(f'    {repr(translated_id)}')
        Translations for 'animals':
            '0:Fido, nice=Maybe?'
        Translations for 'people':
            '0:Bob, nice=Maybe?'

        Since we didn't give an explicit `default_fmt_placeholders`, the regular `fmt` is used instead. Formats can be
        plain strings, in which case translation will never explicitly fail unless the name itself fails to map and
        :attr:`Mapper.unmapped_values_action <.mapping.Mapper.unmapped_values_action>` is set to
        :attr:`ActionLevel.RAISE <rics.action_level.ActionLevel.RAISE>`.
    """

    def __init__(
        self,
        fetcher: FetcherTypes[NameType, SourceType, IdType] = None,
        fmt: FormatType = "{id}:{name}",
        mapper: Mapper[NameType, SourceType, None] = None,
        default_fmt: FormatType = None,
        default_fmt_placeholders: MakeType[SourceType, str, Any] = None,
        allow_name_inheritance: bool = True,
    ) -> None:
        self._fmt = fmt if isinstance(fmt, Format) else Format(fmt)
        self._default_fmt_placeholders: Optional[InheritedKeysDict[SourceType, str, Any]]
        self._default_fmt_placeholders, self._default_fmt = _handle_default(
            self._fmt, default_fmt, default_fmt_placeholders
        )

        self._cached_tmap: TranslationMap[NameType, SourceType, IdType] = TranslationMap({})
        self._fetcher: Fetcher[SourceType, IdType]
        if fetcher is None:
            from .testing import TestFetcher, TestMapper

            self._fetcher = TestFetcher([])  # No explicit sources
            if mapper:  # pragma: no cover
                warnings.warn(
                    f"Mapper instance {mapper} given; consider creating a TestFetcher([sources..])-instance manually.",
                    UserWarning,
                    stacklevel=2,
                )
            else:
                mapper = TestMapper()  # type: ignore
            warnings.warn(
                "No fetcher given. Translation data will be automatically generated.",
                UserWarning,
                stacklevel=2,
            )
        elif isinstance(fetcher, Fetcher):
            self._fetcher = fetcher
        elif isinstance(fetcher, dict):
            self._cached_tmap = self._to_translation_map(
                {source: PlaceholderTranslations.make(source, pht) for source, pht in fetcher.items()}
            )
        elif isinstance(fetcher, TranslationMap):
            tmap = fetcher.copy()
            tmap.fmt = self._fmt
            tmap.default_fmt = self._default_fmt
            tmap.default_fmt_placeholders = self._default_fmt_placeholders  # type: ignore
            self._cached_tmap = tmap
        else:
            raise TypeError(type(fetcher))  # pragma: no cover

        self._mapper: Mapper[NameType, SourceType, None] = mapper or Mapper()
        self._mapper.logger = logging.getLogger(__package__).getChild("mapping").getChild("name-to-source")

        # Misc config
        self._allow_name_inheritance = allow_name_inheritance

        self._config_metadata: Optional[ConfigMetadata] = None

    @classmethod
    def from_config(
        cls,
        path: PathLikeType,
        extra_fetchers: Iterable[PathLikeType] = (),
        clazz: Union[str, Type["Translator[NameType, SourceType, IdType]"]] = None,
    ) -> "Translator[NameType, SourceType, IdType]":
        """Create a ``Translator`` from TOML inputs.

        Args:
            path: Path to the main TOML configuration file.
            extra_fetchers: Paths to fetching configuration TOML files. If multiple fetchers are defined, they are
                ranked by input order. If a fetcher defined in the main configuration, it will be prioritized (rank=0).
            clazz: Translator implementation to create. If a string is passed, the class is resolved using
                :func:`~rics.misc.get_by_full_name` if a string is given. Use ``cls`` if ``None``.

        Returns:
            A new ``Translator`` instance with a :attr:`config_metadata` attribute.
        """
        return TranslatorFactory(
            path,
            extra_fetchers,
            clazz or cls,  # TODO: Higher-Kinded TypeVars
        ).create()

    @property
    def config_metadata(self) -> ConfigMetadata:
        """Return :func:`from_config` initialization metadata."""
        if self._config_metadata is None:
            raise ValueError("Not created using Translator.from_config()")  # pragma: no cover
        return self._config_metadata

    def copy(self, share_fetcher: bool = True, **overrides: Any) -> "Translator[NameType, SourceType, IdType]":
        """Make a copy of this ``Translator``.

        Args:
            share_fetcher: If ``True``, the returned instance use the same ``Fetcher``.
            overrides: Keyword arguments to use when instantiating the copy. Options that aren't given will be taken
                from the current instance. See the :class:`Translator` class documentation for possible choices.

        Returns:
            A copy of this ``Translator`` with `overrides` applied.

        Raises:
            NotImplementedError: If ``share_fetcher=False``.
        """
        if not share_fetcher:
            raise NotImplementedError("Fetcher cloning not implemented.")

        kwargs: Dict[str, Any] = {
            "fmt": self._fmt,
            "default_fmt": self._default_fmt,
            "allow_name_inheritance": self._allow_name_inheritance,
            **overrides,
        }

        if "mapper" not in kwargs:  # pragma: no cover
            kwargs["mapper"] = self.mapper.copy()
        if "default_fmt_placeholders" not in kwargs:
            kwargs["default_fmt_placeholders"] = self._default_fmt_placeholders
        if "fetcher" not in kwargs:
            kwargs["fetcher"] = self.fetcher if self.online else self._cached_tmap.copy()

        return type(self)(**kwargs)

    def translate(
        self,
        translatable: Translatable,
        names: NameTypes[NameType] = None,
        ignore_names: Names[NameType] = None,
        inplace: bool = False,
        override_function: UserOverrideFunction[NameType, SourceType, IdType] = None,
        maximal_untranslated_fraction: float = 1.0,
        reverse: bool = False,
        attribute: str = None,
    ) -> Optional[Translatable]:
        """Translate IDs to human-readable strings.

        For an introduction to translation, see the :ref:`translation-primer` page.

        Args:
            translatable: A data structure to translate.
            names: Explicit names to translate. Derive from `translatable` if ``None``.
            ignore_names: Names **not** to translate, or a predicate ``(str) -> bool``.
            inplace: If ``True``, translate in-place and return ``None``.
            override_function: A callable ``(name, fetcher.sources, ids) -> Source | None``. See :meth:`.Mapper.apply`
                for details.
            maximal_untranslated_fraction: The maximum fraction of IDs for which translation may fail before an error is
                raised. 1=disabled. Ignored in `reverse` mode.
            reverse: If ``True``, perform translations back to IDs. Offline mode only.
            attribute: If given, translate ``translatable.attribute`` instead. If ``inplace=False``, the translated
                attribute will be assigned to `translatable` using
                ``setattr(translatable, attribute, <translated-attribute>)``.

        Returns:
            A translated copy of `translatable` if ``inplace=False``, otherwise ``None``.

        Raises:
            UntranslatableTypeError: If ``type(translatable)`` cannot be translated.
            AttributeError: If `names` are not given and cannot be derived from `translatable`.
            MappingError: If any required (explicitly given) names fail to map to a source.
            MappingError: If name-to-source mapping is ambiguous.
            ValueError: If `maximal_untranslated_fraction` is not a valid fraction.
            TooManyFailedTranslationsError: If translation fails for more than `maximal_untranslated_fraction` of IDs.
            ConnectionStatusError: If ``reverse=True`` while the ``Translator`` is online.
            UserMappingError: If `override_function` returns a source which is not known, and
                ``self.mapper.unknown_user_override_action != 'ignore'``.

        See Also:
            🔑 This is a key event method. See :ref:`key-events` for details.
        """  # noqa: DAR101 darglint is bugged here
        start = perf_counter()

        should_emit_key_event = LOGGER.isEnabledFor(logging.INFO) if self.online else LOGGER.isEnabledFor(logging.DEBUG)
        if should_emit_key_event:
            event_key = f"{self.__class__.__name__.upper()}.TRANSLATE"

            type_name = _resolve_type_name(translatable, attribute)
            name_info = f"Will be derived based on {type_name!r}" if names is None else str(names)
            if ignore_names is not None:
                name_info += f", excluding those given by {ignore_names=}"

            sources = self.sources  # Ensures that the fetcher is warmed up; good for log organization.
            LOGGER.log(
                level=logging.INFO if self.online else logging.DEBUG,
                msg=f"Begin translation of {type_name!r} using {sources=}. Names to translate: {name_info}.",
                extra=dict(
                    event_key=event_key,
                    event_stage="ENTER",
                    event_title=f"{event_key}.ENTER",
                    translatable_type=type_name,
                    names=names,
                    ignore_names=tname(ignore_names, prefix_classname=True) if callable(ignore_names) else ignore_names,
                    inplace=inplace,
                    sources=sources,
                    fmt=repr(self._fmt),
                    maximal_untranslated_fraction=maximal_untranslated_fraction,
                    attribute=attribute,
                    reverse=reverse,
                    online=self.online,
                ),
            )

        if self.online and reverse:  # pragma: no cover
            raise ConnectionStatusError("Reverse translation cannot be performed online.")

        if not (0.0 <= maximal_untranslated_fraction <= 1):  # pragma: no cover
            raise ValueError(f"Argument {maximal_untranslated_fraction=} is not a valid fraction")

        if attribute:
            obj, translatable = translatable, getattr(translatable, attribute)
        else:
            obj = None

        translation_map, names_to_translate = self._get_updated_tmap(
            translatable,
            names,
            ignore_names=ignore_names,
            override_function=override_function,
            parent=obj if (obj is not None and self._allow_name_inheritance) else None,
        )
        if not translation_map:
            return None if inplace else translatable  # pragma: no cover

        translatable_io = resolve_io(translatable)
        if LOGGER.isEnabledFor(logging.DEBUG) or maximal_untranslated_fraction < 1:
            self._verify_translations(
                translatable, names_to_translate, translation_map, translatable_io, maximal_untranslated_fraction
            )

        translation_map.reverse_mode = reverse
        try:
            ans = translatable_io.insert(translatable, names=names_to_translate, tmap=translation_map, copy=not inplace)
        finally:
            translation_map.reverse_mode = False

        if attribute and not inplace and ans is not None:
            setattr(obj, attribute, ans)
            # Hacky special handling for e.g. pandas.Index
            if hasattr(ans, "name") and hasattr(translatable, "name"):  # pragma: no cover
                ans.name = translatable.name
            ans = obj

        if should_emit_key_event:
            execution_time = perf_counter() - start
            inplace_info = f"Values in {type_name!r} have been replaced " if inplace else "Returning a translated copy"
            LOGGER.log(
                level=logging.INFO if self.online else logging.DEBUG,
                msg=f"Finished translation of {type_name!r} in {format_seconds(execution_time)}. {inplace_info} since {inplace=}.",
                extra=dict(
                    event_key=event_key,
                    event_stage="EXIT",
                    event_title=f"{event_key}.EXIT",
                    execution_time=execution_time,
                    translatable_type=type_name,
                    names=names_to_translate,
                    name_to_source_mapping=translation_map.name_to_source,
                    inplace=inplace,
                    fmt=repr(translation_map.fmt),
                    maximal_untranslated_fraction=maximal_untranslated_fraction,
                    attribute=attribute,
                    reverse=reverse,
                    online=self.online,
                ),
            )
        else:
            pass  # pragma: no cover

        return ans

    def map(  # noqa: A003
        self,
        translatable: Translatable,
        names: NameTypes[NameType] = None,
        ignore_names: Names[NameType] = None,
        override_function: UserOverrideFunction[NameType, SourceType, IdType] = None,
    ) -> Optional[DirectionalMapping[NameType, SourceType]]:
        """Map names to translation sources.

        Args:
            translatable: A data structure to map names for.
            names: Explicit names to translate. Derive from `translatable` if ``None``.
            ignore_names: Names **not** to translate, or a predicate ``(str) -> bool``.
            override_function: A callable ``(name, fetcher.sources, ids) -> Source | None``. See
                :meth:`Mapper.apply <.mapping.Mapper.apply>` for details.

        Returns:
            A mapping of names to translation sources. Returns ``None`` if mapping failed.

        Raises:
            AttributeError: If `names` are not given and cannot be derived from `translatable`.
            MappingError: If any required (explicitly given) names fail to map to a source.
            MappingError: If name-to-source mapping is ambiguous.
            UserMappingError: If `override_function` returns a source which is not known, and
                ``self.mapper.unknown_user_override_action != 'ignore'``.

        See Also:
            🔑 This is a key event method. See :ref:`key-events` for details.
        """
        return self._map_inner(translatable, names, ignore_names=ignore_names, override_function=override_function)

    def map_scores(
        self,
        translatable: Translatable,
        names: NameTypes[NameType] = None,
        ignore_names: Names[NameType] = None,
        override_function: UserOverrideFunction[NameType, SourceType, IdType] = None,
    ) -> pd.DataFrame:
        """Returns raw match scores for name-to-source mapping. See :meth:`map` for details."""
        names_to_translate = self._resolve_names(translatable, names, ignore_names)
        return self.mapper.compute_scores(names_to_translate, self.sources, override_function=override_function)

    @property
    def sources(self) -> List[SourceType]:
        """A list of known Source names, such as ``cities`` or ``languages`` (from :attr:`fetcher` or :attr:`cache`)."""
        return list(self.placeholders)

    @property
    def placeholders(self) -> Dict[SourceType, List[str]]:
        """Placeholders for all known Source names, such as ``id`` or ``name`` (from :attr:`fetcher` or :attr:`cache`).

        These are the (possibly unmapped) placeholders that may be used for translation.

        Returns:
            A dict ``{source: [placeholders..]}``.
        """  # noqa: DAR202
        return self._fetcher.placeholders if self.online else self._cached_tmap.placeholders

    def _map_inner(
        self,
        translatable: Translatable,
        names: NameTypes[NameType] = None,
        ignore_names: Names[NameType] = None,
        override_function: UserOverrideFunction[NameType, SourceType, IdType] = None,
        parent: Translatable = None,
    ) -> Optional[DirectionalMapping[NameType, SourceType]]:
        start = perf_counter()
        names_to_translate = self._resolve_names(translatable, names, ignore_names, parent)

        if LOGGER.isEnabledFor(logging.DEBUG):
            event_key = f"{self.__class__.__name__.upper()}.MAP"
            type_name = _resolve_type_name(translatable)
            sources = self.sources
            LOGGER.debug(
                f"Begin name-to-source mapping of names={names_to_translate} in {type_name} against {sources=}.",
                extra=dict(
                    event_key=event_key,
                    event_stage="ENTER",
                    event_title=f"{event_key}.ENTER",
                    translatable_type=type_name,
                    values=names_to_translate,
                    candidates=sources,
                    context=None,
                ),
            )
        name_to_source = self.mapper.apply(names_to_translate, self.sources, override_function=override_function)
        if LOGGER.isEnabledFor(logging.DEBUG):
            execution_time = perf_counter() - start
            LOGGER.debug(
                f"Finished name-to-source mapping of names={names_to_translate} in {type_name} against {sources=}:"
                f" {name_to_source.left_to_right}.",
                extra=dict(
                    event_key=event_key,
                    event_stage="EXIT",
                    event_title=f"{event_key}.EXIT",
                    execution_time=execution_time,
                    translatable_type=type_name,
                    mapping=name_to_source.left_to_right,
                    context=None,
                ),
            )

        unmapped = set() if names is None else set(as_list(names)).difference(name_to_source.left)
        if unmapped or not name_to_source.left:
            params_info = (
                f"could not be mapped to sources={self.sources}. "
                "Additional parameters: ("
                f"{ignore_names=}, "
                f"override_function={tname(override_function)}, "
                f"parent={tname(parent)}"
                ")"
            )

            if names is None:
                derived_names = self._extract_from_attribute_or_parent(translatable, parent)
                msg = f"Translation aborted; none of the derived names {derived_names} {params_info}."
                warnings.warn(msg, MappingWarning, stacklevel=2)
                LOGGER.warning(msg)
                return None
            elif unmapped:
                # Fail if any of the explicitly given names fail to map to a source.
                msg = f"Required names {unmapped} {params_info}."
                LOGGER.error(msg)
                raise MappingError(msg)

        if name_to_source.cardinality.many_right:  # pragma: no cover
            for value, candidates in name_to_source.left_to_right.items():
                if len(candidates) > 1:
                    raise MappingError(
                        f"Name-to-source mapping {name_to_source.left_to_right} is ambiguous; {value} -> {candidates}."
                        f"\nHint: Choose a different cardinality such that Mapper.cardinality.many_right is False."
                    )
        return name_to_source

    def fetch(
        self,
        translatable: Translatable,
        name_to_source: DirectionalMapping[NameType, SourceType],
        data_structure_io: Type[DataStructureIO] = None,
    ) -> TranslationMap[NameType, SourceType, IdType]:
        """Fetch translations.

        Args:
            translatable: A data structure to translate.
            name_to_source: Mappings of names in `translatable` to the known :attr:`sources`.
            data_structure_io: Static namespace used to extract IDs from `translatable`.

        Returns:
            A ``TranslationMap``.

        Raises:
            ConnectionStatusError: If disconnected from the fetcher, i.e. not :attr:`online`.
        """
        ids_to_fetch = self._get_ids_to_fetch(
            name_to_source, translatable, data_structure_io or resolve_io(translatable)
        )
        source_translations = self._fetch(ids_to_fetch)
        return self._to_translation_map(source_translations)

    @property
    def online(self) -> bool:
        """Return connectivity status. If ``False``, no new translations may be fetched."""
        return hasattr(self, "_fetcher")

    @property
    def fetcher(self) -> Fetcher[SourceType, IdType]:
        """Return the ``Fetcher`` instance used to retrieve translations."""
        if not self.online:
            raise ConnectionStatusError("Cannot fetch new translations.")  # pragma: no cover

        return self._fetcher

    @property
    def mapper(self) -> Mapper[NameType, SourceType, None]:
        """Return the ``Mapper`` instance used for name-to-source binding."""
        return self._mapper

    @property
    def cache(self) -> TranslationMap[NameType, SourceType, IdType]:
        """Return a ``TranslationMap`` of cached translations."""
        return self._cached_tmap

    @classmethod
    def load_persistent_instance(
        cls,
        cache_dir: PathLikeType,
        config_path: PathLikeType,
        extra_fetchers: Iterable[PathLikeType] = (),
        max_age: Union[str, pd.Timedelta, timedelta] = "12h",
        clazz: Union[str, Type["Translator[NameType, SourceType, IdType]"]] = None,
    ) -> "Translator[NameType, SourceType, IdType]":
        """Load or create a persistent :attr:`~.Fetcher.fetch_all`-instance.

        Instances are created, stored and loaded as determined by a metadata file located in the given `cache_dir`. A
        new ``Translator`` will be created if:

        * There is no `'metadata'` file, or
        * the original ``Translator`` is too old (see `max_age`), or
        * the current configuration -- as defined by ``(config_path, extra_fetchers, clazz)`` -- has changed in such a
          way that it is no longer equivalent configuration used to create the original ``Translator``. For details, see
          :class:`ConfigMetadata`.

        .. warning:: This method is **not** thread safe.

        Args:
            cache_dir: Root directory where the cached translator and associated metadata is stored.
            config_path: Path to the main TOML configuration file.
            extra_fetchers: Paths to fetching configuration TOML files. If multiple fetchers are defined, they are
                ranked by input order. If a fetcher defined in the main configuration, it will be prioritized (rank=0).
            max_age: The maximum age of the cached ``Translator`` before it must be recreated. Pass ``max_age='0d'`` to
                force recreation.
            clazz: Translator implementation to create. If a string is passed, the class is resolved using
                :func:`~rics.misc.get_by_full_name` if a string is given. Use ``cls`` if ``None``.

        Returns:
            A new or cached ``Translator`` instance with a :attr:`config_metadata` attribute.

        See Also:
             The :meth:`from_config` method, which will initialize the Translator using `path`, `extra_fetchers`, and
                `clazz` if the cached instance is outdated.
        """
        path = Path(str(config_path))
        cache_dir = Path(str(cache_dir)).expanduser().absolute()
        cache_dir.mkdir(parents=True, exist_ok=True)

        metadata_path = cache_dir.joinpath("metadata.json")
        cache_path = cache_dir.joinpath("translator.pkl")

        extra_fetcher_paths: List[str] = list(map(str, extra_fetchers))

        metadata = ConfigMetadata.from_toml_paths(
            str(path),
            extra_fetcher_paths,
            clazz=TranslatorFactory.resolve_class(clazz),
        )
        if metadata.use_cached(metadata_path, pd.Timedelta(max_age).to_pytimedelta()):
            return cls.restore(cache_path)
        else:
            ans: Translator[NameType, SourceType, IdType] = cls.from_config(path, extra_fetcher_paths, clazz)
            ans.store(path=cache_path, delete_fetcher=True)
            metadata_path.write_text(ans.config_metadata.to_json())
            return ans

    @classmethod
    def restore(cls, path: PathLikeType) -> "Translator[NameType, SourceType, IdType]":
        """Restore a serialized ``Translator``.

        Args:
            path: Path to a serialized ``Translator``.

        Returns:
            A ``Translator``.

        Raises:
            TypeError: If the object at `path` is not a ``Translator`` or a subtype thereof.

        See Also:
            The :meth:`Translator.store` method.
        """
        import pickle  # noqa: S403

        full_path = Path(str(path)).expanduser()
        with full_path.open("rb") as f:
            ans = pickle.load(f)  # noqa: S301

        if not isinstance(ans, cls):  # pragma: no cover
            raise TypeError(f"Serialized object at at '{full_path}' is a {type(ans)}, not {cls}.")

        if LOGGER.isEnabledFor(logging.DEBUG):
            extra = "" if ans._config_metadata is None else f" with {ans.config_metadata}"
            LOGGER.debug(f"Deserialized {ans}{extra}.")

        return ans

    def store(
        self,
        translatable: Translatable = None,
        names: NameTypes[NameType] = None,
        ignore_names: Names[NameType] = None,
        delete_fetcher: bool = True,
        path: PathLikeType = None,
    ) -> "Translator[NameType, SourceType, IdType]":  # noqa: DAR401  false positive
        """Retrieve and store translations in memory.

        Args:
            translatable: Data from which IDs to fetch will be extracted. Fetch all IDs if ``None``.
            names: Explicit names to translate. Derive from `translatable` if ``None``.
            ignore_names: Names **not** to translate, or a predicate ``(str) -> bool``.
            delete_fetcher: If ``True``, invoke :meth:`.Fetcher.close` and delete the fetcher after retrieving data. The
                ``Translator`` will still function, but new data cannot be retrieved.
            path: If given, serialize the ``Translator`` to disk after retrieving data.

        Returns:
            Self, for chained assignment.

        Raises:
            ForbiddenOperationError: If :meth:`.Fetcher.fetch_all` is disabled and ``translatable=None``.
            MappingError: If :meth:`map` fails (only when `translatable` is given).

        Notes:
            The ``Translator`` is guaranteed to be :func:`~rics.misc.serializable` once offline. Fetchers often
            aren't as they require things like database connections to function.

        See Also:
            The :meth:`Translator.restore` method.
        """
        start = perf_counter()
        if translatable is None:
            source_translations: SourcePlaceholderTranslations[SourceType] = self._fetch(None)
            translation_map = self._to_translation_map(source_translations)
        else:
            maybe_none, _ = self._get_updated_tmap(translatable, names, ignore_names=ignore_names, force_fetch=True)
            if maybe_none is None:
                raise MappingError("No values in the translatable were mapped. Cannot store translations.")
            translation_map = maybe_none  # mypy, would be cleaner to just use translation map..
            if LOGGER.isEnabledFor(logging.DEBUG):
                not_fetched = set(self.fetcher.sources).difference(translation_map.sources)
                LOGGER.debug(f"Available sources {not_fetched} were not fetched.")

        if delete_fetcher:  # pragma: no cover
            self.fetcher.close()
            del self._fetcher

        self._cached_tmap = translation_map

        message = f"Created {self} in {format_perf_counter(start)}."
        if path:
            import os
            import pickle  # noqa: S403

            path = Path(str(path)).expanduser()
            path.parent.mkdir(parents=True, exist_ok=True)
            with open(path, "wb") as f:
                pickle.dump(self, f)

            mb_size = os.path.getsize(path) / 2**20
            message += f" Serialized {mb_size:.2g} MiB at path='{path}'."

        LOGGER.info(message)
        return self

    def _get_updated_tmap(
        self,
        translatable: Translatable,
        names: NameTypes[NameType] = None,
        ignore_names: Names[NameType] = None,
        override_function: UserOverrideFunction[NameType, SourceType, IdType] = None,
        force_fetch: bool = False,
        parent: Translatable = None,
    ) -> Tuple[Optional[TranslationMap[NameType, SourceType, IdType]], List[NameType]]:
        """Get an updated translation map.  # noqa

        Setting ``force_fetch=True`` will ignore the cached translation if there is one.

        Steps:
            1. Resolve which data structure IO to use, fail if not found.
            2. Resolve name-to-source mappings. If none are found, return ``None``.
            3. Create a new translation map, or update the cached one.

        See the :meth:`translate`-method for more detailed documentation.
        """
        translatable_io = resolve_io(translatable)  # Fail fast if untranslatable type

        name_to_source = self._map_inner(translatable, names, ignore_names, override_function, parent)
        if name_to_source is None:
            # Nothing to translate.
            return None, []  # pragma: no cover

        translation_map = (
            self.fetch(translatable, name_to_source, translatable_io) if force_fetch or not self.cache else self.cache
        )

        n2s = name_to_source.flatten()
        translation_map.name_to_source = n2s  # Update
        return translation_map, list(n2s)

    def _get_ids_to_fetch(
        self,
        name_to_source: DirectionalMapping[NameType, SourceType],
        translatable: Translatable,
        dio: Type[DataStructureIO],
    ) -> List[IdsToFetch[SourceType, IdType]]:
        # Aggregate and remove duplicates.
        source_to_ids: Dict[SourceType, Set[IdType]] = {source: set() for source in name_to_source.right}
        n2s = name_to_source.flatten()  # Will fail if sources are ambiguous.

        float_names: List[NameType] = []
        num_coerced = 0
        for name, ids in dio.extract(translatable, list(n2s)).items():
            if len(ids) == 0:
                continue

            if isinstance(ids[0], float):
                float_names.append(name)
                # Float IDs aren't officially supported, but is common when using Pandas since int types cannot be NaN.
                # This is sometimes a problem for the built-in set (see https://github.com/numpy/numpy/issues/9358), and
                # for several database drivers.
                arr = numpy.unique(ids)
                keep_mask = ~numpy.isnan(arr)
                num_coerced += keep_mask.sum()  # Somewhat inaccurate; includes repeat IDs from other names
                source_to_ids[n2s[name]].update(arr[keep_mask].astype(int, copy=False))
            else:
                source_to_ids[n2s[name]].update(ids)  # type: ignore[arg-type]

        if num_coerced > 100 and self.online:  # pragma: no cover
            warnings.warn(
                f"To ensure proper fetcher operation, {num_coerced} float-type IDs have been coerced to integers. "
                f"Enforcing supported data types for IDs (str and int) in your {tname(translatable)}-data may improve "
                f"performance. Affected names ({len(float_names)}): {float_names}."
                "\nHint: Going offline will suppress this warning.",
                stacklevel=3,
            )

        return [IdsToFetch(source, ids) for source, ids in source_to_ids.items()]

    def _fetch(
        self, ids_to_fetch: Optional[List[IdsToFetch[SourceType, IdType]]]
    ) -> SourcePlaceholderTranslations[SourceType]:
        placeholders = self._fmt.placeholders
        required = self._fmt.required_placeholders

        if self._default_fmt and ID in self._default_fmt.placeholders and ID not in placeholders:
            # Ensure that default translations can always use the ID
            placeholders = placeholders + (ID,)
            required = required + (ID,)

        return (
            self.fetcher.fetch_all(placeholders, required)
            if ids_to_fetch is None
            else self.fetcher.fetch(ids_to_fetch, placeholders, required)
        )

    def _to_translation_map(
        self, source_translations: SourcePlaceholderTranslations[SourceType]
    ) -> TranslationMap[NameType, SourceType, IdType]:
        return TranslationMap(
            source_translations,
            fmt=self._fmt,
            default_fmt=self._default_fmt,
            default_fmt_placeholders=self._default_fmt_placeholders,
        )

    @staticmethod
    def _verify_translations(
        translatable: Translatable,
        names_to_translate: List[NameType],
        translation_map: TranslationMap[NameType, SourceType, IdType],
        translatable_io: Type[DataStructureIO],
        maximal_untranslated_fraction: float,
    ) -> None:
        copied_map = translation_map.copy()
        # TODO: Remove the ignores when https://github.com/python/mypy/issues/3004 (5+ years old..) is fixed.
        copied_map.fmt = "found"  # type: ignore
        copied_map.default_fmt = ""  # type: ignore
        ans = translatable_io.insert(translatable, names=names_to_translate, tmap=copied_map, copy=True)
        extracted = translatable_io.extract(ans, names=names_to_translate)

        total_untranslated = 0
        for name, translations in extracted.items():
            num = sum(t == "" for t in translations)
            total_untranslated += num

            if num == 0:
                continue

            frac = num / len(translations)
            if LOGGER.isEnabledFor(logging.DEBUG) or frac > maximal_untranslated_fraction:
                source = translation_map.name_to_source[name]
                msg = f"Failed to translate {num}/{len(translations)} ({frac:.3%}) of IDs for {name=} using {source=}."
                LOGGER.debug(msg)
                if frac > maximal_untranslated_fraction:
                    raise TooManyFailedTranslationsError(
                        msg + f" Limit: maximal_untranslated_fraction={maximal_untranslated_fraction:.3%}"
                    )

        if total_untranslated or LOGGER.isEnabledFor(logging.DEBUG):
            n_ids = sum(map(len, extracted.values()))
            LOGGER.log(
                logging.WARNING if maximal_untranslated_fraction < 1 else logging.DEBUG,
                f"Failed to translate {total_untranslated}/{n_ids} ({total_untranslated / n_ids:.3%}) "
                f"of IDs extracted from {len(extracted)} different names.",
            )

    def __repr__(self) -> str:
        more = f"fetcher={self.fetcher}" if self.online else f"cache={self.cache}"

        online = self.online
        return f"{tname(self)}({online=}: {more})"

    def _resolve_names(
        self,
        translatable: Translatable,
        names: NameTypes[NameType] = None,
        ignored_names: Names[NameType] = None,
        parent: Translatable = None,  # This isn't correct; should be different typevars.
    ) -> List[NameType]:
        if names is None:
            names = self._extract_from_attribute_or_parent(translatable, parent)
        else:
            names = as_list(names)
        ignored_names = ignored_names if callable(ignored_names) else set(as_list(ignored_names))
        return self._resolve_names_inner(names, ignored_names)

    def _extract_from_attribute_or_parent(
        self, translatable: Translatable, parent: Optional[Translatable]
    ) -> List[NameType]:
        if parent is None:
            names = self._extract_from_attribute(translatable)
        else:
            try:
                names = self._extract_from_attribute(translatable)
            except AttributeError:
                names = self._extract_from_attribute(parent)
                LOGGER.debug(
                    f"Using {names=} from parent of type {tname(parent)} for child of type {tname(translatable)}"
                )
        return names

    @classmethod
    def _extract_from_attribute(cls, translatable: Translatable) -> List[NameType]:
        no_use_keys = False
        for attr_name in _NAME_ATTRIBUTES:
            if attr_name == "keys" and no_use_keys:
                continue  # Pandas Series have keys, but should not be used.

            if hasattr(translatable, attr_name):
                attr = getattr(translatable, attr_name)
                if attr is None:
                    no_use_keys = True
                else:
                    return as_list(attr() if callable(attr) else attr)

        raise AttributeError(
            "Must pass 'names' since no valid name could be found for data of type "
            f"'{tname(translatable)}'."
            f" Attributes checked: {_NAME_ATTRIBUTES}."
        )

    @classmethod
    def _resolve_names_inner(
        cls, names: List[NameType], ignored_names: Union[NamesPredicate[NameType], Set[NameType]]
    ) -> List[NameType]:
        predicate = _IgnoredNamesPredicate(ignored_names).apply
        names_to_translate = list(filter(predicate, names))
        if not names_to_translate and names:
            warnings.warn(
                f"No names left to translate. Ignored names: {ignored_names}, explicit names: {names}.", stacklevel=3
            )
        return names_to_translate


class _IgnoredNamesPredicate(Generic[NameType]):
    def __init__(self, ignored_names: Union[NamesPredicate[NameType], Set[NameType]]) -> None:
        self._func: NamesPredicate[NameType]
        if callable(ignored_names):
            self._func = ignored_names
        else:
            self._func = ignored_names.__contains__

    def apply(self, name: NameType) -> bool:
        return not self._func(name)


def _handle_default(
    fmt: Format,
    default_fmt: Optional[FormatType],
    default_fmt_placeholders: Optional[MakeType[SourceType, str, Any]],
) -> Tuple[Optional[InheritedKeysDict[SourceType, str, Any]], Optional[Format]]:  # pragma: no cover
    if default_fmt is None and default_fmt_placeholders is None:
        return None, None

    dt: Optional[InheritedKeysDict[SourceType, str, Any]] = None

    if isinstance(default_fmt_placeholders, InheritedKeysDict):
        dt = default_fmt_placeholders
    elif isinstance(default_fmt_placeholders, dict):
        dt = InheritedKeysDict.make(default_fmt_placeholders)

    dfmt: Optional[Format]
    if default_fmt is None:
        dfmt = None
    elif isinstance(default_fmt, str):
        dfmt = Format(default_fmt)
    else:
        dfmt = default_fmt

    if dt is not None and dfmt is None:
        # Force a default format if default translations are given
        dfmt = fmt

    if dfmt is not None:
        dt = dt or InheritedKeysDict()

    return dt, dfmt


def _resolve_type_name(translatable: Translatable, attribute: str = None) -> str:
    type_name = tname(translatable, prefix_classname=True)
    if attribute:
        type_name = f"{type_name}.{attribute}"
    return type_name



"""Supporting functions for implementations."""
from typing import Collection, List

from ..types import IdType, SourceType
from .types import FetchInstruction as _FetchInstruction


def select_placeholders(instr: _FetchInstruction[SourceType, IdType], known_placeholders: Collection[str]) -> List[str]:
    """Select from a subset of known placeholders.

    Args:
        instr: Instruction object with placeholders.
        known_placeholders: A collection of known placeholders.

    Returns:
        As many known placeholders from `instr` as possible.
    """
    return list(known_placeholders if instr.fetch_all else filter(known_placeholders.__contains__, instr.placeholders))



"""Types related to translation fetching."""

from dataclasses import dataclass as _dataclass
from typing import Generic as _Generic, Iterable, List, Optional, Set

from ..offline.types import PlaceholdersTuple
from ..types import IdType, SourceType


@_dataclass(frozen=True)
class IdsToFetch(_Generic[SourceType, IdType]):
    """A source and the IDs to fetch from it."""

    source: SourceType
    """Where to fetch from."""
    ids: Iterable[IdType]
    """Unique IDs to fetch translations for.``"""


@_dataclass(frozen=True)
class FetchInstruction(_Generic[SourceType, IdType]):
    """Instructions passed from an ``AbstractFetcher`` to an implementation."""

    source: SourceType
    """Where to fetch from."""
    placeholders: PlaceholdersTuple
    """All desired placeholders in preferred order."""
    required: Set[str]
    """Placeholders that must be included in the response."""
    ids: Optional[List[IdType]]
    """Unique IDs to fetch translations for. Fetch as much as possible if ``None``."""

    @property
    def fetch_all(self) -> bool:
        """Flag indicated that as much data as possible should be retrieved."""
        return self.ids is None



"""Functions which perform heuristics for score functions.

See Also:
    The :class:`~.HeuristicScore` class.
"""
from __future__ import annotations

import re
from typing import Any, Iterable, List, Optional, Set, Tuple, Union

from . import filter_functions as ff
from .types import ContextType

VERBOSE: bool = False
"""If ``True`` enable optional DEBUG-level log messages on each heuristic function invocation.

Notes:
    Not all functions have verbose messages.
"""


def like_database_table(
    name: str,
    candidates: Iterable[str],
    context: Optional[ContextType],
) -> Tuple[str, List[str]]:
    """Try to make `value` look like the name of a database table."""

    def apply(s: str) -> str:
        s = s.lower().replace("_", "").replace(".", "")
        s = s[: -len("id")] if s.endswith("id") else s
        s = s if s.endswith("s") else s + "s"
        return s

    return apply(name), list(map(apply, candidates))


def short_circuit_to_value(
    value: str,
    candidates: Iterable[str],
    context: Optional[str],
    regex: Union[str, re.Pattern[str]],
    target: str,
) -> Set[str]:
    """Short circuit candidates which match a given `regex` a given to-value.

    Args:
        value: A value to map.
        candidates: Candidates for `value`.
        context: Context in which the function is being called.
        regex: A pattern in `candidates` which should trigger forced short-circuit matching.
        target: The target value. If ``value != target``, an empty set is always returned.

    Returns:
        Candidates which match `regex`, or an empty set.

    Notes:
        This is technically a filter function and may be used as such.
    """
    return (
        set()
        if value != target
        else ff.require_regex_match(
            value,
            candidates,
            context,
            regex=regex,
            where="candidate",
            purpose=f"short-circuiting to value-{target=}",
        )
    )


def short_circuit_to_candidate(
    value: str,
    candidates: Iterable[str],
    context: Optional[str],
    regex: Union[str, re.Pattern[str]],
    target: str,
) -> Set[str]:
    """Short circuit candidates which match a given `regex` to a given to-candidate.

    Args:
        value: A value to map.
        candidates: Candidates for `value`.
        context: Context in which the function is being called.
        regex: A pattern in `candidates` which should trigger forced short-circuit matching.
        target: A target candidate. Must be present in `candidates`, or empty set is always returned.

    Returns:
        Candidates which match `regex`, or an empty set.

    Notes:
        This is technically a filter function and may be used as such.
    """
    return (
        set()
        if target not in candidates
        else ff.require_regex_match(
            value,
            [target],
            context,
            regex=regex,
            where="name",
            purpose=f"short-circuiting to candidate-{target=}",
        )
    )


def force_lower_case(
    value: str, candidates: Iterable[str], context: Optional[ContextType]
) -> Tuple[str, Iterable[str]]:
    """Force lower-case in `value` and `candidates`."""
    return value.lower(), list(map(str.lower, candidates))


def value_fstring_alias(
    value: str,
    candidates: Iterable[str],
    context: Optional[ContextType],
    fstring: str,
    for_value: str = None,
    **kwargs: Any,
) -> Tuple[str, Iterable[str]]:
    """Return a value formatted by `fstring`.

    Args:
        value: An element to find matches for.
        candidates: Potential matches for `value`. Not used (returned as given).
        context: Context in which the function is being called.
        fstring: The format string to use. Can use `value` and `context` as placeholders.
        for_value: If given, apply only if ``value == for_value``. When `if_value_equals` is given, `fstring` arguments
            which do not use the `value` as a placeholder key are permitted.
        **kwargs: Additional keyword placeholders in `fstring`.

    Returns:
        A tuple ``(formatted_value, candidates)``.

    Raises:
        ValueError: If `fstring` does not contain a placeholder `'value'` and `for_value` is not given.
    """
    if not for_value and "{value}" not in fstring:
        # No longer a function of the value.
        raise ValueError(
            f"Invalid {fstring=} passed to value_fstring_alias(); does not contain {{value}}. "
            "To allow, the 'for_value' parameter must be given as well."
        )

    if for_value and value != for_value:
        return value, candidates

    return fstring.format(value=value, context=context, **kwargs), candidates


def candidate_fstring_alias(
    value: str,
    candidates: Iterable[str],
    context: Optional[ContextType],
    fstring: str,
    **kwargs: Any,
) -> Tuple[str, Iterable[str]]:
    """Return candidates formatted by `fstring`.

    Args:
        value: An element to find matches for. Not used (returned as given).
        candidates: Potential matches for `value`.
        context: Context in which the function is being called.
        fstring: The format string to use. Can use `value`, `context`, and elements of `candidates` as placeholders.
        **kwargs: Additional keyword placeholders in `fstring`.

    Returns:
        A tuple ``(value, formatted_candidates)``.

    Raises:
        ValueError: If `fstring` does not contain a placeholder `'candidate'`.
    """
    if "{candidate}" not in fstring:
        raise ValueError(f"Invalid {fstring=} passed to candidate_fstring_alias(); does not contain {{candidate}}.")

    return value, map(lambda c: fstring.format(value=value, candidate=c, context=context, **kwargs), candidates)



import json
import logging
from hashlib import sha256
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Tuple, Type

from ._base_metadata import BaseMetadata
from ._load_toml import load_toml_file

LOGGER = logging.getLogger(__package__).getChild("Translator").getChild("config")

if TYPE_CHECKING:
    from . import Translator  # noqa: F401


class ConfigMetadata(BaseMetadata):
    """Metadata pertaining to how a ``Translator`` instance was initialized from TOML configuration.

    Configs are equivalent iff:

        - They have the same top-level dependency versions, and
        - Use the same fully qualified class name, and
        - The main configuration is equal after parsing, and
        - They have the same number of auxiliary (`"extra"`) fetcher configurations, and
        - All auxiliary fetcher configurations are equal after parsing.

    Args:
        main: Absolute path and fingerprint of the main translation configuration."
        extra_fetchers: Absolute path and fingerprint of configuration files for auxiliary fetchers.
        clazz: String representation of the class type.
    """

    def __init__(
        self,
        main: Tuple[Path, str],
        extra_fetchers: Tuple[Tuple[Path, str], ...],
        clazz: str,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.main = main
        self.extra_fetchers = extra_fetchers
        self.clazz = clazz

    def _serialize(self, to_json: Dict[str, Any]) -> Dict[str, Any]:
        kwargs = dict(
            main=tuple(map(str, to_json.pop("main"))),
            extra_fetchers=[tuple(map(str, t)) for t in to_json.pop("extra_fetchers")],
        )
        kwargs["class"] = to_json.pop("clazz")
        return kwargs

    @classmethod
    def _deserialize(cls, from_json: Dict[str, Any]) -> Dict[str, Any]:
        def to_path_tuple(args: List[str]) -> Tuple[Path, str]:
            return Path(args[0]), args[1]

        return dict(
            main=to_path_tuple(from_json.pop("main")),
            extra_fetchers=tuple(map(to_path_tuple, from_json.pop("extra_fetchers"))),
            clazz=from_json.pop("class"),
        )

    def _log_reject(self, msg: str) -> None:
        LOGGER.info(msg.format(slug="Create new Translator"))  # noqa: G001

    def _log_accept(self, msg: str) -> None:
        LOGGER.info(msg.format(kind="Translator"))  # noqa: G001

    def _is_equivalent(self, other: BaseMetadata) -> str:  # pragma: no cover
        assert isinstance(other, ConfigMetadata)  # noqa: S101

        if self.clazz != other.clazz:
            return f"Class not equal. Expected '{self.clazz}', but got '{other.clazz}'"

        if self.main[1] != other.main[1]:
            return f"Main configuration changed. Expected fingerprint {self.main[1]}, but got {other.main[1]}"

        if len(self.extra_fetchers) != len(other.extra_fetchers):
            return (
                f"Number of auxiliary fetchers changed. Expected {len(self.extra_fetchers)}"
                f" fetchers but got {len(other.extra_fetchers)}"
            )

        for i, (path, fingerprint) in enumerate(self.extra_fetchers):
            _, other_fingerprint = other.extra_fetchers[i]
            if fingerprint != other_fingerprint:
                return (
                    f"Configuration changed for auxiliary fetcher #{i} at {path}. "
                    f"Expected fingerprint {fingerprint}, but got {other_fingerprint}"
                )

        return ""

    @staticmethod
    def from_toml_paths(
        path: str,
        extra_fetchers: List[str],
        clazz: Type["Translator[Any, Any, Any]"],
    ) -> "ConfigMetadata":
        """Convenience function for creating ``ConfigMetadata`` instances."""
        return ConfigMetadata(
            main=_create_path_tuple(path),
            extra_fetchers=tuple(map(_create_path_tuple, extra_fetchers)),
            clazz=clazz.__module__ + "." + clazz.__qualname__,  # Fully qualified name
        )


def _create_path_tuple(str_path: str) -> Tuple[Path, str]:
    p = Path(str_path).expanduser().absolute()
    content = load_toml_file(p)
    return p, sha256(json.dumps(content).encode()).hexdigest()
